# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0


import numpy as np
import pytest

from power_grid_model import ComponentType, DatasetType, LoadGenType, initialize_array
from power_grid_model._utils import compatibility_convert_row_columnar_dataset, is_columnar
from power_grid_model.enum import ComponentAttributeFilterOptions
from power_grid_model.validation import validate_batch_data
from power_grid_model.validation.errors import MultiComponentNotUniqueError, NotBetweenOrAtError, NotBooleanError


@pytest.fixture
def original_input_data() -> dict[str, np.ndarray]:
    node = initialize_array(DatasetType.input, ComponentType.node, 4)
    node["id"] = [1, 2, 3, 4]
    node["u_rated"] = 10.5e3
    line = initialize_array(DatasetType.input, ComponentType.line, 4)

    line["id"] = [5, 6, 7, 8]
    line["from_node"] = [1, 2, 3, 1]
    line["to_node"] = [2, 3, 1, 2]
    line["from_status"] = 0
    line["to_status"] = 0
    line["r1"] = 1.0
    line["x1"] = 2.0
    line["c1"] = 3.0
    line["tan1"] = 4.0
    line["i_n"] = 5.0

    asym_load = initialize_array(DatasetType.input, ComponentType.asym_load, 2)
    asym_load["id"] = [9, 10]
    asym_load["node"] = [1, 2]
    asym_load["status"] = [1, 1]
    asym_load["type"] = [LoadGenType.const_power, LoadGenType.const_power]
    asym_load["p_specified"] = [[11e6, 12e6, 13e6], [21e6, 22e6, 23e6]]
    asym_load["q_specified"] = [[11e5, 12e5, 13e5], [21e5, 22e5, 23e5]]

    return {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.asym_load: asym_load,
    }


@pytest.fixture
def original_input_data_columnar_all(original_input_data):
    return compatibility_convert_row_columnar_dataset(
        original_input_data,
        ComponentAttributeFilterOptions.everything,
        DatasetType.input,
    )


@pytest.fixture
def original_input_data_columnar_relevant(original_input_data):
    return compatibility_convert_row_columnar_dataset(
        original_input_data, ComponentAttributeFilterOptions.relevant, DatasetType.input
    )


@pytest.fixture(
    params=[
        "original_input_data",
        "original_input_data_columnar_all",
        "original_input_data_columnar_relevant",
    ]
)
def input_data(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def original_batch_data() -> dict[str, np.ndarray]:
    line = initialize_array(DatasetType.update, ComponentType.line, (3, 2))
    line["id"] = [[5, 6], [6, 7], [7, 5]]
    line["from_status"] = [[1, 1], [1, 1], [1, 1]]

    # Add batch for asym_load, which has 2-D array for p_specified
    asym_load = initialize_array(DatasetType.update, ComponentType.asym_load, (3, 2))
    asym_load["id"] = [[9, 10], [9, 10], [9, 10]]

    return {"line": line, "asym_load": asym_load}


@pytest.fixture
def original_batch_data_columnar_all(original_batch_data):
    return compatibility_convert_row_columnar_dataset(
        original_batch_data,
        ComponentAttributeFilterOptions.everything,
        DatasetType.update,
    )


@pytest.fixture
def original_batch_data_columnar_relevant(original_batch_data):
    return compatibility_convert_row_columnar_dataset(
        original_batch_data,
        ComponentAttributeFilterOptions.relevant,
        DatasetType.update,
    )


@pytest.fixture(
    params=[
        "original_batch_data",
        "original_batch_data_columnar_all",
        "original_batch_data_columnar_relevant",
    ]
)
def batch_data(request):
    return request.getfixturevalue(request.param)


def test_validate_batch_data(input_data, batch_data):
    errors = validate_batch_data(input_data, batch_data)
    assert not errors


def test_validate_batch_data_input_error(input_data, batch_data):
    if is_columnar(input_data["node"]):
        input_data["node"]["id"][-1] = 123
        input_data["line"]["id"][-1] = 123
    else:
        input_data["node"][-1]["id"] = 123
        input_data["line"][-1]["id"] = 123
    errors = validate_batch_data(input_data, batch_data)
    assert len(errors) == 3
    assert [MultiComponentNotUniqueError([("line", "id"), ("node", "id")], [("line", 123), ("node", 123)])] == errors[0]
    assert [MultiComponentNotUniqueError([("line", "id"), ("node", "id")], [("line", 123), ("node", 123)])] == errors[1]
    assert [MultiComponentNotUniqueError([("line", "id"), ("node", "id")], [("line", 123), ("node", 123)])] == errors[2]


def test_validate_batch_data_update_error(input_data, batch_data):
    batch_data["line"]["from_status"] = np.array([[12, 34], [0, -128], [56, 78]])
    errors = validate_batch_data(input_data, batch_data)
    assert len(errors) == 2
    assert 1 not in errors
    assert len(errors[0]) == 1
    assert len(errors[2]) == 1
    assert errors[0] == [NotBooleanError("line", "from_status", [5, 6])]
    assert errors[2] == [NotBooleanError("line", "from_status", [5, 7])]


def test_validate_batch_data_transformer_tap_nom():
    node_input = initialize_array(DatasetType.input, ComponentType.node, 2)
    node_input["id"] = [1, 2]
    node_input["u_rated"] = [400, 10500]

    transformer_input = initialize_array(DatasetType.input, ComponentType.transformer, 1)
    transformer_input["id"] = 3
    transformer_input["from_node"] = 1
    transformer_input["to_node"] = 2
    transformer_input["from_status"] = 1
    transformer_input["to_status"] = 1
    transformer_input["u1"] = 10750
    transformer_input["u2"] = 420
    transformer_input["sn"] = 400000
    transformer_input["uk"] = 0.04
    transformer_input["pk"] = 3750
    transformer_input["i0"] = 0.002230015414744929
    transformer_input["p0"] = 515
    transformer_input["winding_from"] = 2
    transformer_input["winding_to"] = 1
    transformer_input["clock"] = 5
    transformer_input["tap_side"] = 0
    transformer_input["tap_pos"] = 1
    transformer_input["tap_min"] = 1
    transformer_input["tap_max"] = 5
    transformer_input["tap_size"] = 250
    test_input_data = {
        ComponentType.node: node_input,
        ComponentType.transformer: transformer_input,
    }
    test_update_data = {ComponentType.sym_load: initialize_array(DatasetType.update, ComponentType.sym_load, (2, 0))}
    result = validate_batch_data(test_input_data, test_update_data)
    assert result is not None
    assert len(result) == test_update_data[ComponentType.sym_load].shape[0]
    assert len(result[0]) == test_input_data[ComponentType.transformer].shape[0]
    assert len(result[1]) == test_input_data[ComponentType.transformer].shape[0]

    error = NotBetweenOrAtError(ComponentType.transformer, "tap_nom", [3], ("tap_min", "tap_max"))
    assert result == {0: [error], 1: [error]}



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
This file contains helper functions for library-internal use only.

Disclaimer!

We do not officially support this functionality and may remove features in this library at any given time!
"""

from copy import deepcopy
from typing import Sequence, cast

import numpy as np

from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.error_handling import VALIDATOR_MSG
from power_grid_model._core.power_grid_meta import initialize_array, power_grid_meta_data
from power_grid_model.data_types import (
    BatchColumn,
    BatchComponentData,
    BatchDataset,
    BatchList,
    ComponentData,
    Dataset,
    DenseBatchArray,
    DenseBatchColumnarData,
    DenseBatchData,
    IndexPointer,
    PythonDataset,
    SingleArray,
    SingleColumn,
    SingleColumnarData,
    SingleComponentData,
    SingleDataset,
    SinglePythonDataset,
    SparseBatchData,
)
from power_grid_model.enum import ComponentAttributeFilterOptions
from power_grid_model.errors import PowerGridError
from power_grid_model.typing import ComponentAttributeMapping, _ComponentAttributeMappingDict


def is_nan(data) -> bool:
    """
    Determine if the data point is valid
    Args:
        data: a single scaler or numpy array

    Returns:
        True if all the data points are invalid
        False otherwise
    """
    nan_func = {
        np.dtype("f8"): lambda x: np.all(np.isnan(x)),
        np.dtype("i4"): lambda x: np.all(x == np.iinfo("i4").min),
        np.dtype("i1"): lambda x: np.all(x == np.iinfo("i1").min),
    }
    return bool(nan_func[data.dtype](data))


def convert_batch_dataset_to_batch_list(batch_data: BatchDataset, dataset_type: DatasetType | None = None) -> BatchList:
    """
    Convert batch datasets to a list of individual batches

    Args:
        batch_data: a batch dataset for power-grid-model
        dataset_type: type of dataset

    Returns:
        A list of individual batches
    """

    # If the batch data is empty, return an empty list
    if len(batch_data) == 0:
        return []

    n_batches = get_and_verify_batch_sizes(batch_data=batch_data, dataset_type=dataset_type)

    # Initialize an empty list with dictionaries
    # Note that [{}] * n_batches would result in n copies of the same dict.
    list_data: BatchList = [{} for _ in range(n_batches)]

    # While the number of batches must be the same for each component, the structure (2d numpy array or indptr/data)
    # doesn't have to be. Therefore, we'll check the structure for each component and copy the data accordingly.
    for component, data in batch_data.items():
        component_data_checks(data, component)
        component_batches: Sequence[SingleComponentData]
        if is_sparse(data):
            component_batches = split_sparse_batch_data_in_batches(cast(SparseBatchData, data), component)
        else:
            component_batches = split_dense_batch_data_in_batches(cast(SingleComponentData, data), batch_size=n_batches)
        for i, batch in enumerate(component_batches):
            if (isinstance(batch, dict) and batch) or (isinstance(batch, np.ndarray) and batch.size > 0):
                list_data[i][component] = batch
    return list_data


def get_and_verify_batch_sizes(batch_data: BatchDataset, dataset_type: DatasetType | None = None) -> int:
    """
    Determine the number of batches for each component and verify that each component has the same number of batches

    Args:
        batch_data: a batch dataset for power-grid-model
        dataset_type: type of dataset

    Returns:
        The number of batches
    """

    if dataset_type is None and any(is_columnar(v) and not is_sparse(v) for v in batch_data.values()):
        dataset_type = get_dataset_type(batch_data)

    n_batch_size = 0
    checked_components: list[ComponentType] = []
    for component, data in batch_data.items():
        n_component_batch_size = get_batch_size(data, dataset_type, component)
        if checked_components and n_component_batch_size != n_batch_size:
            if len(checked_components) == 1:
                checked_components_str = f"'{checked_components.pop()}'"
            else:
                str_checked_components = [str(component) for component in checked_components]
                checked_components_str = "/".join(sorted(str_checked_components))
            raise ValueError(
                f"Inconsistent number of batches in batch data. "
                f"Component '{component}' contains {n_component_batch_size} batches, "
                f"while {checked_components_str} contained {n_batch_size} batches."
            )
        n_batch_size = n_component_batch_size
        checked_components.append(component)
    return n_batch_size


def get_batch_size(
    batch_data: BatchComponentData, dataset_type: DatasetType | None = None, component: ComponentType | None = None
) -> int:
    """
    Determine the number of batches and verify the data structure while we're at it. Note only batch data is supported.
    Note: SingleColumnarData would get treated as batch by this function.

    Args:
        batch_data: a batch array for power-grid-model
        dataset_type: type of dataset
        component: name of component

    Raises:
        ValueError: when the type for data_filter is incorrect

    Returns:
        The number of batches
    """
    component_data_checks(batch_data)
    if is_sparse(batch_data):
        indptr = batch_data["indptr"]
        return indptr.size - 1

    if not is_columnar(batch_data):
        sym_array = batch_data
    else:
        batch_data = cast(DenseBatchColumnarData, batch_data)
        if component is None or dataset_type is None:
            raise ValueError("Cannot deduce batch size for given columnar data without a dataset type or component")
        sym_attributes, _ = _get_sym_or_asym_attributes(dataset_type, component)
        for attribute, array in batch_data.items():
            if attribute in sym_attributes:
                break
            if array.ndim == 1:
                raise TypeError("Incorrect dimension present in batch data.")
            if array.ndim == 2:
                return 1
            return array.shape[0]
        sym_array = next(iter(batch_data.values()))

    sym_array = cast(DenseBatchArray | BatchColumn, sym_array)
    if sym_array.ndim == 3:
        raise TypeError("Incorrect dimension present in batch data.")
    if sym_array.ndim == 1:
        return 1
    return sym_array.shape[0]


def _get_sym_or_asym_attributes(dataset_type: DatasetType, component: ComponentType):
    """Segregate into symmetric of asymmetric attribute.
    The asymmetric attribute is per phase value and of extra dimension.

    Args:
        dataset_type (DatasetType): dataset type
        component (ComponentType): component name

    Returns:
        symmetrical and asymmetrical attributes
    """
    asym_attributes = set()
    sym_attributes = set()
    for meta_dataset_type, dataset_meta in power_grid_meta_data.items():
        if dataset_type != meta_dataset_type:
            continue
        for component_name_meta, component_meta in dataset_meta.items():
            if component != component_name_meta:
                continue
            if component_meta.dtype.names is None:
                raise ValueError("No attributes available in meta")
            for attribute in component_meta.dtype.names:
                if component_meta.dtype[attribute].shape == (3,):
                    asym_attributes.add(attribute)
                if component_meta.dtype[attribute].shape == ():
                    sym_attributes.add(attribute)
    return sym_attributes, asym_attributes


def _split_numpy_array_in_batches(
    data: DenseBatchArray | SingleArray | SingleColumn | BatchColumn,
) -> list[SingleArray] | list[SingleColumn]:
    """
    Split a single dense numpy array into one or more batches

    Args:
        data: A 1D or 2D Numpy structured array. A 1D array is a single table / batch, a 2D array is a batch per table.

    Returns:
        A list with a single numpy structured array per batch
    """
    if data.ndim == 1:
        return [data]
    if data.ndim in [2, 3]:
        return [data[i, ...] for i in range(data.shape[0])]
    raise ValueError("Dimension of the component data is invalid.")


def split_dense_batch_data_in_batches(
    data: SingleComponentData | DenseBatchData, batch_size: int
) -> list[SingleComponentData]:
    """
    Split a single dense numpy array into one or more batches

    Args:
        data: A 1D or 2D Numpy structured array. A 1D array is a single table / batch, a 2D array is a batch per table.
        batch_size: size of batch

    Returns:
        A list with a single component data per scenario
    """
    if isinstance(data, np.ndarray):
        return cast(list[SingleComponentData], _split_numpy_array_in_batches(data))

    scenarios_per_attribute = {
        attribute: _split_numpy_array_in_batches(attribute_data) for attribute, attribute_data in data.items()
    }

    return [
        {attribute: scenarios_per_attribute[attribute][scenario] for attribute, attribute_data in data.items()}
        for scenario in range(batch_size)
    ]


def split_sparse_batch_data_in_batches(
    batch_data: SparseBatchData, component: ComponentType
) -> list[SingleComponentData]:
    """
    Split a single numpy array representing, a compressed sparse structure, into one or more batches

    Args:
        batch_data: Sparse batch data
        component: The name of the component to which the data belongs, only used for errors.

    Returns:
        A list with a single numpy structured array per batch
    """
    for key in ["indptr", "data"]:
        if key not in batch_data:
            raise KeyError(
                f"Missing '{key}' in sparse batch data for '{component}' "
                "(expected a python dictionary containing two keys: 'indptr' and 'data')."
            )

    data = batch_data["data"]
    indptr = batch_data["indptr"]

    def _split_buffer(buffer: np.ndarray, scenario: int) -> SingleArray:
        if not isinstance(buffer, np.ndarray) or buffer.ndim != 1:
            raise TypeError(
                f"Invalid data type {type(buffer).__name__} in sparse batch data for '{component}' "
                "(should be a 1D Numpy structured array (i.e. a single 'table'))."
            )

        if not isinstance(indptr, np.ndarray) or indptr.ndim != 1 or not np.issubdtype(indptr.dtype, np.integer):
            raise TypeError(
                f"Invalid indptr data type {type(indptr).__name__} in batch data for '{component}' "
                "(should be a 1D Numpy array (i.e. a single 'list'), "
                "containing indices (i.e. integers))."
            )

        if indptr[0] != 0 or indptr[-1] != len(buffer) or indptr[scenario] > indptr[scenario + 1]:
            raise TypeError(
                f"Invalid indptr in batch data for '{component}' "
                f"(should start with 0, end with the number of objects ({len(buffer)}) "
                "and be monotonic increasing)."
            )

        return buffer[indptr[scenario] : indptr[scenario + 1]]

    def _get_scenario(scenario: int) -> SingleComponentData:
        if isinstance(data, dict):
            return {attribute: _split_buffer(attribute_data, scenario) for attribute, attribute_data in data.items()}
        return _split_buffer(data, scenario)

    return [_get_scenario(i) for i in range(len(indptr) - 1)]


def convert_dataset_to_python_dataset(data: Dataset) -> PythonDataset:
    """
    Convert internal numpy arrays to native python data
      If an attribute is not available (NaN value), it will not be exported.

    Args:
        data: A single or batch dataset for power-grid-model
    Returns:
        A python dict for single dataset
        A python list for batch dataset
    """

    # Check if the dataset is a single dataset or batch dataset
    # It is batch dataset if it is 2D array or a indptr/data structure
    is_batch: bool | None = None
    for component, array in data.items():
        is_dense_batch = isinstance(array, np.ndarray) and array.ndim == 2
        is_sparse_batch = isinstance(array, dict) and "indptr" in array and "data" in array
        if is_batch is not None and is_batch != (is_dense_batch or is_sparse_batch):
            raise ValueError(
                f"Mixed {'' if is_batch else 'non-'}batch data "
                f"with {'non-' if is_batch else ''}batch data ({component})."
            )
        is_batch = is_dense_batch or is_sparse_batch

    # If it is a batch, convert the batch data to a list of batches, then convert each batch individually.
    if is_batch:
        # We have established that this is batch data, so let's tell the type checker that this is a BatchDataset
        data = cast(BatchDataset, data)
        list_data = convert_batch_dataset_to_batch_list(data)
        return [convert_single_dataset_to_python_single_dataset(data=x) for x in list_data]

    # We have established that this is not batch data, so let's tell the type checker that this is a SingleDataset
    data = cast(SingleDataset, data)
    return convert_single_dataset_to_python_single_dataset(data=data)


def convert_single_dataset_to_python_single_dataset(
    data: SingleDataset,
) -> SinglePythonDataset:
    """
    Convert internal numpy arrays to native python data
    If an attribute is not available (NaN value), it will not be exported.

    Args:
        data: A single dataset for power-grid-model

    Returns:
        A python dict for single dataset
    """

    # Convert each numpy array to a list of objects, which contains only the non-NaN attributes:
    # For example: {"node": [{"id": 0, ...}, {"id": 1, ...}], "line": [{"id": 2, ...}]}
    def _convert_component(objects: SingleComponentData):
        # This should be a single data set
        if not isinstance(objects, np.ndarray) or objects.ndim != 1:
            raise ValueError("Invalid data format")

        return [
            {attribute: obj[attribute].tolist() for attribute in objects.dtype.names if not is_nan(obj[attribute])}
            for obj in objects
        ]

    return {component: _convert_component(objects) for component, objects in data.items()}


def compatibility_convert_row_columnar_dataset(
    data: Dataset,
    data_filter: ComponentAttributeMapping,
    dataset_type: DatasetType,
    available_components: list[ComponentType] | None = None,
) -> Dataset:
    """Temporary function to transform row, column or mixed based datasets to a full row or column based dataset as per
    the data_filter. The purpose of this function is to mimic columnar data and transform back to row data without any
    memory footprint benefits.
    Note: Copies are made on a per-component basis; if a component is row based in both the input and the requested
    output, that componened is returned without a copy.

    Args:
        data (Dataset): dataset to convert
        data_filter (ComponentAttributeMapping): desired component and attribute mapping
        dataset_type (DatasetType): type of dataset (e.g., input, update or [sym | asym | sc]_output)
        available_components (list[ComponentType] | None): available components in model

    Returns:
        Dataset: converted dataset
    """
    if available_components is None:
        available_components = list(data.keys())

    processed_data_filter = process_data_filter(dataset_type, data_filter, available_components)

    result_data: Dataset = {}
    for comp_name, attrs in processed_data_filter.items():
        if comp_name not in data:
            continue

        sub_data = _extract_data_from_component_data(data[comp_name])
        converted_sub_data = _convert_data_to_row_or_columnar(
            data=sub_data,
            comp_name=comp_name,
            dataset_type=dataset_type,
            attrs=attrs,
        )

        if is_sparse(data[comp_name]):
            result_data[comp_name] = {"indptr": _extract_indptr(data[comp_name]), "data": converted_sub_data}
        else:
            result_data[comp_name] = converted_sub_data
    return result_data


def _convert_data_to_row_or_columnar(
    data: SingleComponentData,
    comp_name: ComponentType,
    dataset_type: DatasetType,
    attrs: set[str] | list[str] | None | ComponentAttributeFilterOptions,
) -> SingleComponentData:
    """Converts row or columnar component data to row or columnar component data as requested in `attrs`."""
    if attrs is None:
        if not is_columnar(data):
            return data
        data = cast(SingleColumnarData, data)
        output_array = initialize_array(dataset_type, comp_name, next(iter(data.values())).shape)
        for k in data:
            output_array[k] = data[k]
        return output_array

    if isinstance(attrs, (list, set)) and len(attrs) == 0:
        return {}
    if isinstance(attrs, ComponentAttributeFilterOptions):
        names = cast(SingleArray, data).dtype.names if not is_columnar(data) else cast(SingleColumnarData, data).keys()
        return {attr: deepcopy(data[attr]) for attr in names}
    return {attr: deepcopy(data[attr]) for attr in attrs}


def process_data_filter(
    dataset_type: DatasetType,
    data_filter: ComponentAttributeMapping,
    available_components: list[ComponentType],
) -> _ComponentAttributeMappingDict:
    """Checks valid type for data_filter. Also checks for any invalid component names and attribute names.

    Args:
        dataset_type (DatasetType): the type of output that the user will see (as per the calculation options)
        data_filter (ComponentAttributeMapping):  data_filter provided by user
        available_components (list[ComponentType]):  all components available in model instance or data

    Returns:
        _ComponentAttributeMappingDict: processed data_filter in a dictionary
    """
    if data_filter is None:
        processed_data_filter: _ComponentAttributeMappingDict = {ComponentType[k]: None for k in available_components}
    elif isinstance(data_filter, ComponentAttributeFilterOptions):
        processed_data_filter = {ComponentType[k]: data_filter for k in available_components}
    elif isinstance(data_filter, (list, set)):
        processed_data_filter = {ComponentType[k]: None for k in data_filter}
    elif isinstance(data_filter, dict) and all(
        attrs is None or isinstance(attrs, (set, list, ComponentAttributeFilterOptions))
        for attrs in data_filter.values()
    ):
        processed_data_filter = data_filter
    else:
        raise ValueError(f"Invalid filter provided: {data_filter}")

    validate_data_filter(processed_data_filter, dataset_type, available_components)
    return processed_data_filter


def validate_data_filter(
    data_filter: _ComponentAttributeMappingDict,
    dataset_type: DatasetType,
    available_components: list[ComponentType],
) -> None:
    """Raise error if some specified components or attributes are unknown.

    Args:
        data_filter (_ComponentAttributeMappingDict): Processed component to attribtue dictionary
        dataset_type (DatasetType):  Type of dataset
        available_components (list[ComponentType]):  all components available in model instance or data

    Raises:
        ValueError: when the type for data_filter is incorrect
        KeyError: with "unknown component types" for any unknown components
        KeyError: with "unknown attributes" for unknown attribute(s) for a known component
    """
    dataset_meta = power_grid_meta_data[dataset_type]

    for source, components in {
        "data_filter": data_filter.keys(),
        "data": available_components,
    }.items():
        unknown_components = [x for x in components if x not in dataset_meta]
        if unknown_components:
            raise KeyError(f"The following specified component types are unknown:{unknown_components} in {source}")

    unknown_attributes = {}
    for comp_name, attrs in data_filter.items():
        if attrs is None or isinstance(attrs, ComponentAttributeFilterOptions):
            continue

        attr_names = dataset_meta[comp_name].dtype.names
        diff = set(cast(set[str] | list[str], attrs))
        if attr_names is not None:
            diff = diff.difference(attr_names)
        if diff != set():
            unknown_attributes[comp_name] = diff

    if unknown_attributes:
        raise KeyError(f"The following specified attributes are unknown: {unknown_attributes} in data_filter")


def is_sparse(component_data: ComponentData) -> bool:
    """Check if component_data is sparse or dense. Only batch data can be sparse."""
    return isinstance(component_data, dict) and set(component_data.keys()) == {
        "indptr",
        "data",
    }


def is_columnar(component_data: ComponentData) -> bool:
    """Check if component_data is columnar or row based"""
    if is_sparse(component_data):
        return not isinstance(component_data["data"], np.ndarray)
    return not isinstance(component_data, np.ndarray)


def is_nan_or_default(x: np.ndarray) -> np.ndarray:
    """
    Check if elements in the array are NaN or equal to the min of its dtype.

    Args:
        x: A NumPy array to check.

    Returns:
        A boolean NumPy array where each element is True if the corresponding element in x is NaN
        or min of its dtype, and False otherwise.
    """
    if x.dtype == np.float64:
        return np.isnan(x)
    if x.dtype in (np.int32, np.int8):
        return x == np.iinfo(x.dtype).min
    raise TypeError(f"Unsupported data type: {x.dtype}")


def is_nan_or_equivalent(array) -> bool:
    """
    Check if the array contains only nan values or equivalent nan values for specific data types.
    This is the aggregrated version of `is_nan_or_default` for the whole array.

    Args:
        array: The array to check.

    Returns:
        bool: True if the array contains only nan or equivalent nan values, False otherwise.
    """
    return isinstance(array, np.ndarray) and bool(
        (array.dtype == np.float64 and np.isnan(array).all())
        or (array.dtype in (np.int32, np.int8) and np.all(array == np.iinfo(array.dtype).min))
    )


def _check_sparse_dense(component_data: ComponentData, err_msg_suffixed: str) -> ComponentData:
    if is_sparse(component_data):
        indptr = component_data["indptr"]
        if not isinstance(indptr, np.ndarray):
            raise TypeError(err_msg_suffixed.format(f"Invalid indptr type {type(indptr).__name__}. "))
        sub_data = component_data["data"]
    elif isinstance(component_data, dict) and ("indptr" in component_data or "data" in component_data):
        missing_element = "indptr" if "indptr" not in component_data else "data"
        raise KeyError(err_msg_suffixed.format(f"Missing '{missing_element}' in sparse batch data. "))
    else:
        sub_data = component_data
    return sub_data


def _check_columnar_row(sub_data: ComponentData, err_msg_suffixed: str) -> None:
    if is_columnar(sub_data):
        if not isinstance(sub_data, dict):
            raise TypeError(err_msg_suffixed.format(""))
        for attribute, attribute_array in sub_data.items():
            if not isinstance(attribute_array, np.ndarray):
                raise TypeError(err_msg_suffixed.format(f"'{attribute}' attribute. "))
            if attribute_array.ndim not in [1, 2, 3]:
                raise TypeError(err_msg_suffixed.format(f"Invalid dimension: {attribute_array.ndim}"))
    elif not isinstance(sub_data, np.ndarray):
        raise TypeError(err_msg_suffixed.format(f"Invalid data type {type(sub_data).__name__} "))
    elif isinstance(sub_data, np.ndarray) and sub_data.ndim not in [1, 2]:
        raise TypeError(err_msg_suffixed.format(f"Invalid dimension: {sub_data.ndim}. "))


def component_data_checks(component_data: ComponentData, component=None) -> None:
    """Checks if component_data is of ComponentData and raises ValueError if its not"""
    component_name = f"'{component}'" if component is not None else ""
    err_msg = f"Invalid data for {component_name} component. " "{0}"
    err_msg_suffixed = err_msg + "Expecting a 1D/2D Numpy structured array or a dictionary of such."

    sub_data = _check_sparse_dense(component_data, err_msg_suffixed)
    _check_columnar_row(sub_data, err_msg_suffixed)


def _extract_indptr(data: ComponentData) -> IndexPointer:  # pragma: no cover
    """returns indptr and checks if its valid

    Args:
        data (ComponentData): The component data

    Raises:
        TypeError: if indptr is invalid or is not available

    Returns:
        IndexPointer: indptr if present
    """
    if not is_sparse(data):
        raise TypeError("Not sparse data")
    indptr = data["indptr"]
    if not isinstance(indptr, np.ndarray):
        raise TypeError("indptr is not a 1D numpy array")
    if indptr.ndim != 1:
        raise TypeError("indptr is not a 1D numpy array")
    return indptr


def _extract_columnar_data(
    data: ComponentData, is_batch: bool | None = None
) -> SingleColumnarData | DenseBatchColumnarData:  # pragma: no cover
    """returns the contents of the columnar data.

    Args:
        data (ComponentData): component data
        is_batch (bool | None, optional): If given data is batch. Skips batch check if provided None.

    Raises:
        TypeError: if data is not columnar or invalid data

    Returns:
        SingleColumnarData | DenseBatchColumnarData: the contents of columnar data
    """
    not_columnar_data_message = "Expected columnar data"

    if is_batch is not None:
        allowed_dims = [2, 3] if is_batch else [1, 2]
    else:
        allowed_dims = [1, 2, 3]

    sub_data = data["data"] if is_sparse(data) else data

    if not isinstance(sub_data, dict):
        raise TypeError(not_columnar_data_message)
    for attribute, attribute_array in sub_data.items():
        if not isinstance(attribute_array, np.ndarray) or not isinstance(attribute, str):
            raise TypeError(not_columnar_data_message)
        if attribute_array.ndim not in allowed_dims:
            raise TypeError(not_columnar_data_message)
    return cast(SingleColumnarData | DenseBatchColumnarData, sub_data)


def _extract_row_based_data(
    data: ComponentData, is_batch: bool | None = None
) -> SingleArray | DenseBatchArray:  # pragma: no cover
    """returns the contents of the row based data

    Args:
        data (ComponentData): component data
        is_batch (bool | None, optional): If given data is batch. Skips batch check if provided None.

    Raises:
        TypeError: if data is not row based or invalid data

    Returns:
        SingleArray | DenseBatchArray: the contents of row based data
    """
    if is_batch is not None:
        allowed_dims = [2] if is_batch else [1]
    else:
        allowed_dims = [1, 2]

    sub_data = data["data"] if is_sparse(data) else data

    if not isinstance(sub_data, np.ndarray):
        raise TypeError("Expected row based data")
    if sub_data.ndim not in allowed_dims:
        raise TypeError("Expected row based data")
    return sub_data


def _extract_data_from_component_data(data: ComponentData, is_batch: bool | None = None):
    return _extract_columnar_data(data, is_batch) if is_columnar(data) else _extract_row_based_data(data, is_batch)


def _extract_contents_from_data(data: ComponentData):
    return data["data"] if is_sparse(data) else data


def check_indptr_consistency(indptr: IndexPointer, batch_size: int | None, contents_size: int):
    """checks if an indptr is valid. Batch size check is optional.

    Args:
        indptr (IndexPointer): The indptr array
        batch_size (int | None): number of scenarios
        contents_size (int): total number of elements in all scenarios

    Raises:
        ValueError: If indptr is invalid
    """
    if indptr[0] != 0 or indptr[-1] != contents_size:
        raise ValueError(f"indptr should start from zero and end at size of data array. {VALIDATOR_MSG}")
    if np.any(np.diff(indptr) < 0):
        raise ValueError(f"indptr should be increasing. {VALIDATOR_MSG}")
    if batch_size is not None and batch_size != indptr.size - 1:
        raise ValueError(f"Provided batch size must be equal to actual batch size. {VALIDATOR_MSG}")


def get_dataset_type(data: Dataset) -> DatasetType:
    """
    Deduce the dataset type from the provided dataset.

    Args:
        data: the dataset

    Raises:
        ValueError
            if the dataset type cannot be deduced because multiple dataset types match the format
            (probably because the data contained no supported components, e.g. was empty)
        PowerGridError
            if no dataset type matches the format of the data
            (probably because the data contained conflicting data formats)

    Returns:
        The dataset type.
    """
    candidates = set(power_grid_meta_data.keys())

    if all(is_columnar(v) for v in data.values()):
        raise ValueError("The dataset type could not be deduced. At least one component should have row based data.")

    for dataset_type, dataset_metadatas in power_grid_meta_data.items():
        for component, dataset_metadata in dataset_metadatas.items():
            if component not in data or is_columnar(data[component]):
                continue
            component_data = data[component]

            component_dtype = component_data["data"].dtype if is_sparse(component_data) else component_data.dtype
            if component_dtype is not dataset_metadata.dtype:
                candidates.discard(dataset_type)
                break

    if not candidates:
        raise PowerGridError(
            "The dataset type could not be deduced because no type matches the data. "
            "This usually means inconsistent data was provided."
        )
    if len(candidates) > 1:
        raise ValueError("The dataset type could not be deduced because multiple dataset types match the data.")

    return next(iter(candidates))


def get_comp_size(comp_data: SingleColumnarData | SingleArray) -> int:
    """
    Get the number of elements in the comp_data of a single dataset.

    Args:
        comp_data: Columnar or row based data of a single batch

    Returns:
        Number of elements in the component
    """
    if not is_columnar(comp_data):
        return len(comp_data)
    comp_data = cast(SingleColumnarData, comp_data)
    return len(next(iter(comp_data.values())))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0


import numpy as np
import pytest

from power_grid_model import (
    ComponentType,
    DatasetType,
    LoadGenType,
    PowerGridModel,
    initialize_array,
    power_grid_meta_data,
)
from power_grid_model._utils import compatibility_convert_row_columnar_dataset
from power_grid_model.enum import CalculationType, ComponentAttributeFilterOptions
from power_grid_model.validation import assert_valid_batch_data


@pytest.fixture
def input_data_r():
    node = initialize_array(DatasetType.input, ComponentType.node, 3)
    node["id"] = np.array([1, 2, 6])
    node["u_rated"] = [10.5e3, 10.5e3, 10.5e3]

    line = initialize_array(DatasetType.input, ComponentType.line, 3)
    line["id"] = [3, 5, 8]
    line["from_node"] = [1, 2, 1]
    line["to_node"] = [2, 6, 6]
    line["from_status"] = [1, 1, 1]
    line["to_status"] = [1, 1, 1]
    line["r1"] = [0.25, 0.25, 0.25]
    line["x1"] = [0.2, 0.2, 0.2]
    line["c1"] = [10e-6, 10e-6, 10e-6]
    line["tan1"] = [0.0, 0.0, 0.0]
    line["i_n"] = [1000, 1000, 1000]

    sym_load = initialize_array(DatasetType.input, ComponentType.sym_load, 2)
    sym_load["id"] = [4, 7]
    sym_load["node"] = [2, 6]
    sym_load["status"] = [1, 1]
    sym_load["type"] = [LoadGenType.const_power, LoadGenType.const_power]
    sym_load["p_specified"] = [20e6, 10e6]
    sym_load["q_specified"] = [5e6, 2e6]

    source = initialize_array(DatasetType.input, ComponentType.source, 1)
    source["id"] = [10]
    source["node"] = [1]
    source["status"] = [1]
    source["u_ref"] = [1.0]

    return {
        ComponentType.node: node,
        ComponentType.line: line,
        ComponentType.sym_load: sym_load,
        ComponentType.source: source,
    }


@pytest.fixture
def input_data_c(input_data_r):
    return compatibility_convert_row_columnar_dataset(
        input_data_r, ComponentAttributeFilterOptions.relevant, DatasetType.input
    )


@pytest.fixture(params=["input_data_r", "input_data_c"])
def input_data(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def update_sym_load_r():
    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 2))
    sym_load["id"] = [[4], [7]]
    sym_load["p_specified"] = [[30e6], [15e6]]
    return sym_load


@pytest.fixture
def update_sym_load_no_id_r():
    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 2))
    sym_load["p_specified"] = [[30e6, 10e6], [30e6, 15e6]]
    return sym_load


@pytest.fixture
def update_sym_load_c():
    source_attribute_dtypes = power_grid_meta_data[DatasetType.update][ComponentType.sym_load].dtype
    return {
        "id": np.array([[4], [7]], dtype=source_attribute_dtypes["id"]),
        "p_specified": np.array([[30e6], [15e6]], dtype=source_attribute_dtypes["p_specified"]),
    }


@pytest.fixture
def update_sym_load_no_id_c():
    source_attribute_dtypes = power_grid_meta_data[DatasetType.update][ComponentType.sym_load].dtype
    return {
        "p_specified": np.array([[30e6, 10e6], [30e6, 15e6]], dtype=source_attribute_dtypes["p_specified"]),
    }


@pytest.fixture(
    params=[
        "update_sym_load_r",
        "update_sym_load_no_id_r",
        "update_sym_load_c",
        "update_sym_load_no_id_c",
    ]
)
def update_sym_load(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def update_line_r():
    line = initialize_array(DatasetType.update, ComponentType.line, (2, 1))
    line["id"] = [[3], [5]]
    line["from_status"] = [[0], [0]]
    return line


@pytest.fixture
def update_line_no_id_r():
    line = initialize_array(DatasetType.update, ComponentType.line, (2, 3))
    line["from_status"] = [[0, 1, 1], [0, 0, 1]]
    return line


@pytest.fixture
def update_line_c():
    source_attribute_dtypes = power_grid_meta_data[DatasetType.update][ComponentType.line].dtype
    return {
        "id": np.array([[3], [5]], dtype=source_attribute_dtypes["id"]),
        "from_status": np.array([[0], [0]], dtype=source_attribute_dtypes["from_status"]),
    }


@pytest.fixture
def update_line_no_id_c():
    source_attribute_dtypes = power_grid_meta_data[DatasetType.update][ComponentType.line].dtype
    return {
        "from_status": np.array([[0, 1, 1], [0, 0, 1]], dtype=source_attribute_dtypes["from_status"]),
    }


@pytest.fixture(
    params=[
        "update_line_r",
        "update_line_no_id_r",
        "update_line_c",
        "update_line_no_id_c",
    ]
)
def update_line(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def update_data(update_sym_load, update_line):
    return {
        ComponentType.sym_load: update_sym_load,
        ComponentType.line: update_line,
    }


def test_power_flow(input_data, update_data):
    model = PowerGridModel(input_data)
    assert_valid_batch_data(input_data=input_data, update_data=update_data, calculation_type=CalculationType.power_flow)
    model.calculate_power_flow(update_data=update_data)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Data handling
"""

import numpy as np

from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.power_grid_dataset import CConstDataset, CMutableDataset
from power_grid_model._core.power_grid_meta import initialize_array, power_grid_meta_data
from power_grid_model._utils import process_data_filter
from power_grid_model.data_types import Dataset, SingleDataset
from power_grid_model.enum import CalculationType, ComponentAttributeFilterOptions
from power_grid_model.errors import PowerGridUnreachableHitError
from power_grid_model.typing import ComponentAttributeMapping


def get_output_type(*, calculation_type: CalculationType, symmetric: bool) -> DatasetType:
    """
    Get the output type based on the provided arguments.

    Args:
        calculation_type:
            request the output type for a specific calculation type (power_flow, state_estimation, ...)
        symmetric:
            True: three-phase symmetric calculation, even for asymmetric loads/generations
            False: three-phase asymmetric calculation

    Returns:
        the output type that fits the format requested by the output type
    """
    if calculation_type in (CalculationType.power_flow, CalculationType.state_estimation):
        return DatasetType.sym_output if symmetric else DatasetType.asym_output

    if calculation_type == CalculationType.short_circuit:
        return DatasetType.sc_output

    raise NotImplementedError()


def prepare_input_view(input_data: SingleDataset) -> CConstDataset:
    """
    Create a view of the input data in a format compatible with the PGM core libary.

    Args:
        input_data:
            the input data to create the view from

    Returns:
        instance of CConstDataset ready to be fed into C API
    """
    return CConstDataset(input_data, dataset_type=DatasetType.input)


def prepare_update_view(update_data: Dataset) -> CConstDataset:
    """
    Create a view of the update data, or an empty view if not provided, in a format compatible with the PGM core libary.

    Args:
        update_data:
            the update data to create the view from. Defaults to None

    Returns:
        instance of CConstDataset ready to be fed into C API
    """
    return CConstDataset(update_data, dataset_type=DatasetType.update)


def prepare_output_view(output_data: Dataset, output_type: DatasetType) -> CMutableDataset:
    """
    create a view of the output data in a format compatible with the PGM core libary.

    Args:
        output_data:
            the output data to create the view from
        output_type:
            the output type of the output_data

    Returns:
        instance of CMutableDataset ready to be fed into C API
    """
    return CMutableDataset(output_data, dataset_type=output_type)


def create_output_data(
    output_component_types: ComponentAttributeMapping,
    output_type: DatasetType,
    all_component_count: dict[ComponentType, int],
    is_batch: bool,
    batch_size: int,
) -> Dataset:
    """
    Create the output dataset based on component and batch size from the model; and output attributes requested by user.

    Args:
        output_component_types:
            the output components the user seeks to extract
        output_type:
            the type of output that the user will see (as per the calculation options)
        all_component_count:
            the amount of components in the grid (as per the input data)
        is_batch:
            if the dataset is batch
        batch_size:
            the batch size

    Returns:
        Dataset: output dataset
    """
    processed_output_types = process_data_filter(output_type, output_component_types, list(all_component_count.keys()))

    all_component_count = {k: v for k, v in all_component_count.items() if k in processed_output_types}

    # create result dataset
    result_dict: Dataset = {}

    for name, count in all_component_count.items():
        # shape
        if is_batch:
            shape: tuple[int] | tuple[int, int] = (batch_size, count)
        else:
            shape = (count,)

        requested_component = processed_output_types[name]
        dtype = power_grid_meta_data[output_type][name].dtype
        if dtype.names is None:
            raise PowerGridUnreachableHitError
        if requested_component is None:
            result_dict[name] = initialize_array(output_type, name, shape=shape, empty=True)
        elif requested_component in [
            ComponentAttributeFilterOptions.everything,
            ComponentAttributeFilterOptions.relevant,
        ]:
            result_dict[name] = {attr: np.empty(shape, dtype=dtype[attr]) for attr in dtype.names}
        elif isinstance(requested_component, list | set):
            result_dict[name] = {attr: np.empty(shape, dtype=dtype[attr]) for attr in requested_component}

    return result_dict



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pytest

from power_grid_model._core.dataset_definitions import ComponentType, DatasetType
from power_grid_model._core.power_grid_dataset import CConstDataset
from power_grid_model._core.power_grid_meta import power_grid_meta_data
from power_grid_model.errors import PowerGridError


def input_dataset_types():
    return [DatasetType.input]


def update_dataset_types():
    return [DatasetType.update]


def output_dataset_types():
    return [DatasetType.sym_output, DatasetType.asym_output, DatasetType.sc_output]


def all_dataset_types():
    return input_dataset_types() + update_dataset_types() + output_dataset_types()


@pytest.fixture(params=all_dataset_types())
def dataset_type(request):
    return request.param


def test_const_dataset__empty_dataset(dataset_type):
    dataset = CConstDataset(data={}, dataset_type=dataset_type)
    info = dataset.get_info()

    assert info.name() == dataset_type
    assert info.dataset_type() == dataset_type

    assert not info.is_batch()
    assert info.batch_size() == 1
    assert info.n_components() == 0
    assert info.components() == []
    assert info.elements_per_scenario() == {}
    assert info.total_elements() == {}

    with pytest.raises(ValueError):
        CConstDataset(data={})


def test_const_dataset__conflicting_data():
    with pytest.raises(PowerGridError):
        CConstDataset(
            data={
                ComponentType.node: np.zeros(1, dtype=power_grid_meta_data["input"][ComponentType.node]),
                "sym_load": np.zeros(1, dtype=power_grid_meta_data["update"]["sym_load"]),
            }
        )


def test_const_dataset__single_data(dataset_type):
    components = {ComponentType.node: 3, ComponentType.sym_load: 2, ComponentType.asym_load: 4}
    data = {
        component: np.zeros(shape=count, dtype=power_grid_meta_data[dataset_type][component])
        for component, count in components.items()
    }

    dataset = CConstDataset(data, dataset_type)
    info = dataset.get_info()

    assert info.name() == dataset_type
    assert info.dataset_type() == dataset_type

    assert not info.is_batch()
    assert info.batch_size() == 1
    assert info.n_components() == len(components)
    assert info.components() == list(components)
    assert info.elements_per_scenario() == components
    assert info.total_elements() == components


@pytest.mark.parametrize("batch_size", (0, 1, 3))
def test_const_dataset__uniform_batch_data(dataset_type, batch_size):
    components = {ComponentType.node: 3, ComponentType.sym_load: 2, ComponentType.asym_load: 4}
    data = {
        component: np.zeros(shape=(batch_size, count), dtype=power_grid_meta_data[dataset_type][component])
        for component, count in components.items()
    }

    dataset = CConstDataset(data, dataset_type)
    info = dataset.get_info()

    assert info.name() == dataset_type
    assert info.dataset_type() == dataset_type

    assert info.is_batch()
    assert info.batch_size() == batch_size
    assert info.n_components() == len(components)
    assert info.components() == list(components)
    assert info.elements_per_scenario() == components
    assert info.total_elements() == {component: batch_size * count for component, count in components.items()}


def test_const_dataset__sparse_batch_data(dataset_type):
    batch_size = 3
    components = {ComponentType.node: 3, ComponentType.sym_load: 2, ComponentType.asym_load: 4, ComponentType.link: 4}
    data = {
        ComponentType.node: {
            "data": np.zeros(shape=3, dtype=power_grid_meta_data[dataset_type][ComponentType.node]),
            "indptr": np.array([0, 2, 3, 3]),
        },
        ComponentType.sym_load: {
            "data": np.zeros(shape=2, dtype=power_grid_meta_data[dataset_type]["sym_load"]),
            "indptr": np.array([0, 0, 1, 2]),
        },
        ComponentType.asym_load: {
            "data": np.zeros(shape=4, dtype=power_grid_meta_data[dataset_type]["asym_load"]),
            "indptr": np.array([0, 2, 3, 4]),
        },
        ComponentType.link: np.zeros(shape=(batch_size, 4), dtype=power_grid_meta_data[dataset_type]["link"]),
    }

    dataset = CConstDataset(data, dataset_type)
    info = dataset.get_info()

    assert info.name() == dataset_type
    assert info.dataset_type() == dataset_type

    assert info.is_batch()
    assert info.batch_size() == 3
    assert info.n_components() == len(components)
    assert info.components() == list(components)
    assert info.elements_per_scenario() == {
        ComponentType.node: -1,
        ComponentType.sym_load: -1,
        ComponentType.asym_load: -1,
        ComponentType.link: 4,
    }
    assert info.total_elements() == {
        ComponentType.node: 3,
        ComponentType.sym_load: 2,
        ComponentType.asym_load: 4,
        ComponentType.link: batch_size * 4,
    }


def test_const_dataset__mixed_batch_size(dataset_type):
    data = {
        ComponentType.node: np.zeros(shape=(2, 3), dtype=power_grid_meta_data[dataset_type][ComponentType.node]),
        ComponentType.line: np.zeros(shape=(3, 3), dtype=power_grid_meta_data[dataset_type][ComponentType.line]),
    }
    with pytest.raises(ValueError):
        CConstDataset(data, dataset_type)


@pytest.mark.parametrize("bad_indptr", (np.ndarray([0, 1]), np.ndarray([0, 3, 2]), np.ndarray([0, 1, 2, 3, 4])))
def test_const_dataset__bad_sparse_data(dataset_type, bad_indptr):
    data = {
        ComponentType.node: {
            "data": np.zeros(shape=2, dtype=power_grid_meta_data[dataset_type][ComponentType.node]),
            "indptr": bad_indptr,
        },
    }
    with pytest.raises(TypeError):
        CConstDataset(data, dataset_type)


@pytest.mark.parametrize(
    ("dtype", "supported"),
    [
        (power_grid_meta_data[DatasetType.input][ComponentType.node].dtype["id"], True),
        ("<i4", True),
        ("<i8", False),
        ("<i1", False),
        ("<f8", False),
    ],
)
def test_const_dataset__different_dtype(dataset_type, dtype, supported):
    data = {
        ComponentType.node: {
            "id": np.zeros(shape=3, dtype=dtype),
        }
    }
    if supported:
        result = CConstDataset(data, dataset_type)
        assert result.get_info().total_elements() == {ComponentType.node: 3}
    else:
        with pytest.raises(ValueError):
            CConstDataset(data, dataset_type)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import itertools
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from power_grid_model import ComponentType, DatasetType, initialize_array
from power_grid_model._core.dataset_definitions import ComponentType as CT, DatasetType as DT
from power_grid_model._core.power_grid_meta import power_grid_meta_data
from power_grid_model._utils import (
    compatibility_convert_row_columnar_dataset,
    convert_batch_dataset_to_batch_list,
    convert_dataset_to_python_dataset,
    get_and_verify_batch_sizes,
    get_batch_size,
    get_dataset_type,
    is_nan,
    is_sparse,
    process_data_filter,
    split_dense_batch_data_in_batches,
    split_sparse_batch_data_in_batches,
)
from power_grid_model.data_types import BatchDataset, BatchList
from power_grid_model.enum import ComponentAttributeFilterOptions
from power_grid_model.errors import PowerGridError

from .utils import convert_python_to_numpy


@pytest.fixture(name="two_nodes_one_line")
def two_nodes_one_line_fixture():
    return {
        "node": [{"id": 11, "u_rated": 10.5e3}, {"id": 12, "u_rated": 10.5e3}],
        "line": [
            {
                "id": 21,
                "from_node": 11,
                "to_node": 12,
                "from_status": 1,
                "to_status": 1,
                "r1": 0.11,
                "x1": 0.12,
                "c1": 4.1380285203892784e-05,
                "tan1": 0.1076923076923077,
                "i_n": 510.0,
            }
        ],
    }


@pytest.fixture(name="two_nodes_two_lines")
def two_nodes_two_lines_fixture():
    return {
        "node": [{"id": 11, "u_rated": 10.5e3}, {"id": 12, "u_rated": 10.5e3}],
        "line": [
            {
                "id": 21,
                "from_node": 11,
                "to_node": 12,
                "from_status": 1,
                "to_status": 1,
                "r1": 0.11,
                "x1": 0.12,
                "c1": 4.1380285203892784e-05,
                "tan1": 0.1076923076923077,
                "i_n": 510.0,
            },
            {
                "id": 31,
                "from_node": 11,
                "to_node": 12,
                "from_status": 1,
                "to_status": 1,
                "r1": 0.11,
                "x1": 0.12,
                "c1": 4.1380285203892784e-05,
                "tan1": 0.1076923076923077,
                "i_n": 510.0,
            },
        ],
    }


def np_array_int(data):
    return np.array(data, dtype="i4")


def assert_list_of_numpy_arrays_equal(expected, actual):
    assert type(actual) is type(expected)
    assert len(actual) == len(expected)
    for i in range(len(expected)):
        assert isinstance(expected[i], type(actual[i]))
        if isinstance(expected[i], dict):
            assert actual[i].keys() == expected[i].keys()
            for attribute in expected[i]:
                np.testing.assert_array_equal(expected[i][attribute], actual[i][attribute])
        else:
            np.testing.assert_array_equal(expected[i], actual[i])


def assert_list_of_dicts_of_numpy_arrays_equal(expected, actual):
    assert type(actual) is type(expected)
    assert len(actual) == len(expected)
    for i in range(len(expected)):
        assert isinstance(expected[i], type(actual[i]))
        assert actual[i].keys() == expected[i].keys()
        for comp in expected[i]:
            if isinstance(expected[i][comp], dict):
                assert actual[i][comp].keys() == expected[i][comp].keys()
                for attribute in expected[i][comp]:
                    np.testing.assert_array_equal(expected[i][comp][attribute], actual[i][comp][attribute])
            else:
                np.testing.assert_array_equal(expected[i][comp], actual[i][comp])


def test_is_nan():
    single_value = np.array([np.nan])
    assert is_nan(single_value)
    array_f8 = np.array([0.1, 0.2, np.nan], dtype=np.dtype("f8"))
    assert not is_nan(array_f8)
    array_i4 = np.array([10, 2, -(2**31), 40], dtype=np.dtype("i4"))
    assert not is_nan(array_i4)
    array_i1 = np.array([1, 0, -(2**7), 1], dtype=np.dtype("i1"))
    assert not is_nan(array_i1)
    nan_array = np.array([np.nan, np.nan, np.nan])
    assert is_nan(nan_array)


def test_convert_json_to_numpy(two_nodes_one_line, two_nodes_two_lines):
    pgm_data = convert_python_to_numpy(two_nodes_one_line, "input")
    assert len(pgm_data) == 2
    assert len(pgm_data["node"]) == 2
    assert pgm_data["node"][0]["id"] == 11
    assert pgm_data["node"][0]["u_rated"] == 10.5e3
    assert len(pgm_data["line"]) == 1

    json_list = [two_nodes_one_line, two_nodes_two_lines, two_nodes_one_line]
    pgm_data_batch = convert_python_to_numpy(json_list, "input")
    assert pgm_data_batch["node"].shape == (3, 2)
    assert np.allclose(pgm_data_batch["line"]["indptr"], [0, 1, 3, 4])


def test_round_trip_json_numpy_json(two_nodes_one_line, two_nodes_two_lines):
    pgm_data = convert_python_to_numpy(two_nodes_one_line, "input")
    json_dict = convert_dataset_to_python_dataset(pgm_data)
    assert json_dict == two_nodes_one_line

    json_list = [two_nodes_one_line, two_nodes_two_lines, two_nodes_one_line]
    pgm_data_list = convert_python_to_numpy(json_list, "input")
    json_return_list = convert_dataset_to_python_dataset(pgm_data_list)
    assert json_return_list == json_list


def test_convert_batch_to_list_data__zero_batches():
    assert convert_batch_dataset_to_batch_list({}) == []


def test_split_dense_batch_data_in_batches_n1():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    update_data = np.array([(111, 121, 131), (112, 122, 132), (113, 123, 133), (114, 124, 134)], dtype=foo)
    expected = [update_data]
    actual = split_dense_batch_data_in_batches(update_data, 3)
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_split_dense_batch_data_in_batches_n1__columnar():
    update_data = {
        "a": np_array_int([111, 121, 131]),
        "b": np_array_int([112, 122, 132]),
        "c": np_array_int([114, 124, 134]),
    }
    expected = [update_data]
    actual = split_dense_batch_data_in_batches(update_data, 1)
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_split_dense_batch_data_in_batches_n2():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    update_data = np.array(
        [
            [(1111, 1121, 1131), (1112, 1122, 1132), (1113, 1123, 1133), (1114, 1124, 1134)],
            [(2111, 2121, 2131), (2112, 2122, 2132), (2113, 2123, 2133), (2114, 2124, 2134)],
        ],
        dtype=foo,
    )
    expected = [
        np.array([(1111, 1121, 1131), (1112, 1122, 1132), (1113, 1123, 1133), (1114, 1124, 1134)], dtype=foo),
        np.array([(2111, 2121, 2131), (2112, 2122, 2132), (2113, 2123, 2133), (2114, 2124, 2134)], dtype=foo),
    ]
    actual = split_dense_batch_data_in_batches(update_data, 1)
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_split_dense_batch_data_in_batches_n2__columnar():
    update_data = {
        "a": np_array_int([[1111, 1112, 1113, 1114], [2111, 2112, 2113, 2114]]),
        "b": np_array_int([[1121, 1122, 1123, 1124], [2121, 2122, 2123, 2124]]),
        "c": np_array_int([[1131, 1132, 1133, 1134], [2131, 2132, 2133, 2134]]),
    }
    expected = [
        {
            "a": np_array_int([1111, 1112, 1113, 1114]),
            "b": np_array_int([1121, 1122, 1123, 1124]),
            "c": np_array_int([1131, 1132, 1133, 1134]),
        },
        {
            "a": np_array_int([2111, 2112, 2113, 2114]),
            "b": np_array_int([2121, 2122, 2123, 2124]),
            "c": np_array_int([2131, 2132, 2133, 2134]),
        },
    ]
    actual = split_dense_batch_data_in_batches(update_data, 2)
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_split_dense_batch_data_in_batches_wrong_data_dim():
    update_date = np.array([[[[1, 2, 3]]]])
    with pytest.raises(
        ValueError,
        match="Dimension of the component data is invalid",
    ):
        split_dense_batch_data_in_batches(update_date, 1)


def test_normalize_batch_data_structure_n3_sparse():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    update_data = np.array(
        [
            (1111, 1121, 1131),
            (1112, 1122, 1132),
            (1113, 1123, 1133),
            (1114, 1124, 1134),
            (2111, 2121, 2131),
            (2112, 2122, 2132),
            (2113, 2123, 2133),
            (2114, 2124, 2134),
        ],
        dtype=foo,
    )
    indptr = np.array([0, 4, 4, 8])
    expected = [
        np.array([(1111, 1121, 1131), (1112, 1122, 1132), (1113, 1123, 1133), (1114, 1124, 1134)], dtype=foo),
        np.array([], dtype=foo),
        np.array([(2111, 2121, 2131), (2112, 2122, 2132), (2113, 2123, 2133), (2114, 2124, 2134)], dtype=foo),
    ]
    actual = split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr}, component="")
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_normalize_batch_data_structure_n3_sparse__columnar():
    update_data = {
        "a": np_array_int([1111, 1112, 1113, 1114, 2111, 2112, 2113, 2114]),
        "b": np_array_int([1121, 1122, 1123, 1124, 2121, 2122, 2123, 2124]),
        "c": np_array_int([1131, 1132, 1133, 1134, 2131, 2132, 2133, 2134]),
    }
    indptr = np.array([0, 4, 4, 8])
    expected = [
        {
            "a": np_array_int([1111, 1112, 1113, 1114]),
            "b": np_array_int([1121, 1122, 1123, 1124]),
            "c": np_array_int([1131, 1132, 1133, 1134]),
        },
        {"a": np_array_int([]), "b": np_array_int([]), "c": np_array_int([])},
        {
            "a": np_array_int([2111, 2112, 2113, 2114]),
            "b": np_array_int([2121, 2122, 2123, 2124]),
            "c": np_array_int([2131, 2132, 2133, 2134]),
        },
    ]
    actual = split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr}, component="")
    assert_list_of_numpy_arrays_equal(expected, actual)


def test_split_compressed_sparse_structure_in_batches_wrong_data():
    data_1 = [1, 2, 3, 4, 5, 6, 7, 8]  # wrong type
    data_2 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])  # wrong dimension
    indptr = np.array([0, 4, 4, 8])
    with pytest.raises(TypeError, match="Invalid data type list in sparse batch data for 'foo' "):
        split_sparse_batch_data_in_batches(batch_data={"data": data_1, "indptr": indptr}, component="foo")  # type: ignore
    with pytest.raises(TypeError, match="Invalid data type ndarray in sparse batch data for 'bar' "):
        split_sparse_batch_data_in_batches(batch_data={"data": data_2, "indptr": indptr}, component="bar")


def test_split_compressed_sparse_structure_in_batches_wrong_indptr():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    update_data = np.array(
        [
            (1111, 1121, 1131),
            (1112, 1122, 132),
            (1113, 1123, 1133),
            (1114, 1124, 1134),
            (2111, 2121, 2131),
            (2112, 2122, 232),
            (2113, 2123, 2133),
            (2114, 2124, 2134),
        ],
        dtype=foo,
    )
    indptr_1 = [0, 4, 4, 8]  # wrong type
    indptr_2 = np.array([[0, 4], [4, 8]])  # wrong dimension
    indptr_3 = np.array([0.0, 4.0, 4.0, 8.0])  # wrong dtype
    with pytest.raises(TypeError, match="Invalid indptr data type list in batch data for 'foo' "):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_1}, component="foo")  # type: ignore
    with pytest.raises(TypeError, match="Invalid indptr data type ndarray in batch data for 'foo' "):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_2}, component="foo")  # type: ignore
    with pytest.raises(TypeError, match="Invalid indptr data type ndarray in batch data for 'foo' "):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_3}, component="foo")  # type: ignore


def test_split_compressed_sparse_structure_in_batches_wrong_indptr_values():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    update_data = np.array(
        [
            (1111, 1121, 1131),
            (1112, 1122, 132),
            (1113, 1123, 1133),
            (1114, 1124, 1134),
            (2111, 2121, 2131),
            (2112, 2122, 232),
            (2113, 2123, 2133),
            (2114, 2124, 2134),
        ],
        dtype=foo,
    )
    indptr_1 = np.array([1, 4, 4, 8])
    indptr_2 = np.array([0, 4, 4, 9])
    indptr_3 = np.array([0, 4, 3, 8])
    with pytest.raises(
        TypeError,
        match="Invalid indptr in batch data for 'foo' "
        r"\(should start with 0, end with the number of objects \(8\) "
        r"and be monotonic increasing\).",
    ):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_1}, component="foo")
    with pytest.raises(
        TypeError,
        match="Invalid indptr in batch data for 'foo' "
        r"\(should start with 0, end with the number of objects \(8\) "
        r"and be monotonic increasing\).",
    ):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_2}, component="foo")
    with pytest.raises(
        TypeError,
        match="Invalid indptr in batch data for 'foo' "
        r"\(should start with 0, end with the number of objects \(8\) "
        r"and be monotonic increasing\).",
    ):
        split_sparse_batch_data_in_batches(batch_data={"data": update_data, "indptr": indptr_3}, component="foo")


def test_convert_batch_dataset_to_batch_list_one_batch_dense():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    bar = [("x", "i4"), ("y", "i4"), ("z", "i4")]
    update_data: BatchDataset = {
        "foo": np.array([(111, 121, 131), (112, 122, 132), (113, 123, 133), (114, 124, 134)], dtype=foo),
        "bar": np.array([(211, 221, 231), (212, 222, 232), (213, 223, 233), (214, 224, 234)], dtype=bar),
    }
    expected: BatchList = [
        {
            "foo": np.array([(111, 121, 131), (112, 122, 132), (113, 123, 133), (114, 124, 134)], dtype=foo),
            "bar": np.array([(211, 221, 231), (212, 222, 232), (213, 223, 233), (214, 224, 234)], dtype=bar),
        }
    ]
    actual = convert_batch_dataset_to_batch_list(update_data)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


@patch("power_grid_model._utils.get_batch_size")
def test_convert_batch_dataset_to_batch_list_one_batch_dense_columnar(get_batch_size: MagicMock):
    get_batch_size.return_value = 1
    update_data: BatchDataset = {
        "foo": {
            "a": np_array_int([111, 112, 113, 114]),
            "b": np_array_int([121, 122, 123, 124]),
            "c": np_array_int([131, 132, 133, 134]),
        },
        "bar": {
            "a": np_array_int([211, 212, 213, 214]),
            "b": np_array_int([221, 222, 223, 224]),
            "c": np_array_int([231, 232, 233, 234]),
        },
    }
    expected: BatchList = [update_data]
    actual = convert_batch_dataset_to_batch_list(update_data, DT.sym_output)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


def test_convert_batch_dataset_to_batch_list_two_batches_dense():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    bar = [("x", "i4"), ("y", "i4"), ("z", "i4")]
    update_data: BatchDataset = {
        "foo": np.array(
            [
                [(1111, 1121, 1131), (1112, 1122, 132), (1113, 1123, 1133), (1114, 1124, 1134)],
                [(2111, 2121, 2131), (2112, 2122, 232), (2113, 2123, 2133), (2114, 2124, 2134)],
            ],
            dtype=foo,
        ),
        "bar": np.array(
            [
                [(1211, 1221, 1231), (1212, 1222, 232), (1213, 1223, 1233), (1214, 1224, 1234)],
                [(2211, 2221, 2231), (2212, 2222, 232), (2213, 2223, 2233), (2214, 2224, 2234)],
            ],
            dtype=bar,
        ),
    }
    expected: BatchList = [
        {
            "foo": np.array([(1111, 1121, 1131), (1112, 1122, 132), (1113, 1123, 1133), (1114, 1124, 1134)], dtype=foo),
            "bar": np.array([(1211, 1221, 1231), (1212, 1222, 232), (1213, 1223, 1233), (1214, 1224, 1234)], dtype=bar),
        },
        {
            "foo": np.array([(2111, 2121, 2131), (2112, 2122, 232), (2113, 2123, 2133), (2114, 2124, 2134)], dtype=foo),
            "bar": np.array([(2211, 2221, 2231), (2212, 2222, 232), (2213, 2223, 2233), (2214, 2224, 2234)], dtype=bar),
        },
    ]
    actual = convert_batch_dataset_to_batch_list(update_data)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


@patch("power_grid_model._utils.get_batch_size")
def test_convert_batch_dataset_to_batch_list_two_batches_dense__columnar(get_batch_size: MagicMock):
    get_batch_size.return_value = 2
    update_data: BatchDataset = {
        "foo": {
            "a": np_array_int([[1111, 1112, 1113, 1114], [2111, 2112, 2113, 2114]]),
            "b": np_array_int([[1121, 1122, 1123, 1124], [2121, 2122, 2123, 2124]]),
            "c": np_array_int([[1131, 1132, 1133, 1134], [2131, 2132, 2133, 2134]]),
        },
        "bar": {
            "x": np_array_int([[1211, 1212, 1213, 1214], [2211, 2212, 2213, 2214]]),
            "y": np_array_int([[1221, 1222, 1223, 1224], [2221, 2222, 2223, 2224]]),
            "z": np_array_int([[1231, 1232, 1233, 1234], [2231, 2232, 2233, 2234]]),
        },
    }
    expected: BatchList = [
        {
            "foo": {
                "a": np_array_int([1111, 1112, 1113, 1114]),
                "b": np_array_int([1121, 1122, 1123, 1124]),
                "c": np_array_int([1131, 1132, 1133, 1134]),
            },
            "bar": {
                "x": np_array_int([1211, 1212, 1213, 1214]),
                "y": np_array_int([1221, 1222, 1223, 1224]),
                "z": np_array_int([1231, 1232, 1233, 1234]),
            },
        },
        {
            "foo": {
                "a": np_array_int([2111, 2112, 2113, 2114]),
                "b": np_array_int([2121, 2122, 2123, 2124]),
                "c": np_array_int([2131, 2132, 2133, 2134]),
            },
            "bar": {
                "x": np_array_int([2211, 2212, 2213, 2214]),
                "y": np_array_int([2221, 2222, 2223, 2224]),
                "z": np_array_int([2231, 2232, 2233, 2234]),
            },
        },
    ]
    actual = convert_batch_dataset_to_batch_list(update_data, DT.sym_output)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


def test_convert_batch_dataset_to_batch_list_three_batches_sparse():
    foo = [("a", "i4"), ("b", "i4"), ("c", "i4")]
    bar = [("x", "i4"), ("y", "i4"), ("z", "i4")]
    update_data: BatchDataset = {
        "foo": {
            "indptr": np.array([0, 4, 8, 8]),
            "data": np.array(
                [
                    (1111, 1121, 1131),
                    (1112, 1122, 132),
                    (1113, 1123, 1133),
                    (1114, 1124, 1134),
                    (2111, 2121, 2131),
                    (2112, 2122, 232),
                    (2113, 2123, 2133),
                    (2114, 2124, 2134),
                ],
                dtype=foo,
            ),
        },
        "bar": {
            "indptr": np.array([0, 4, 4, 8]),
            "data": np.array(
                [
                    (1211, 1221, 1231),
                    (1212, 1222, 232),
                    (1213, 1223, 1233),
                    (1214, 1224, 1234),
                    (3211, 3221, 3231),
                    (3212, 3222, 332),
                    (3213, 3223, 3233),
                    (3214, 3224, 3234),
                ],
                dtype=bar,
            ),
        },
    }
    expected: BatchList = [
        {
            "foo": np.array([(1111, 1121, 1131), (1112, 1122, 132), (1113, 1123, 1133), (1114, 1124, 1134)], dtype=foo),
            "bar": np.array([(1211, 1221, 1231), (1212, 1222, 232), (1213, 1223, 1233), (1214, 1224, 1234)], dtype=bar),
        },
        {
            "foo": np.array([(2111, 2121, 2131), (2112, 2122, 232), (2113, 2123, 2133), (2114, 2124, 2134)], dtype=foo),
        },
        {"bar": np.array([(3211, 3221, 3231), (3212, 3222, 332), (3213, 3223, 3233), (3214, 3224, 3234)], dtype=bar)},
    ]
    actual = convert_batch_dataset_to_batch_list(update_data)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


def test_convert_batch_dataset_to_batch_list_three_batches_sparse__columnar():
    update_data: BatchDataset = {
        "foo": {
            "indptr": np.array([0, 4, 8, 8]),
            "data": {
                "a": np_array_int([1111, 1112, 1113, 1114, 2111, 2112, 2113, 2114]),
                "b": np_array_int([1121, 1122, 1123, 1124, 2121, 2122, 2123, 2124]),
                "c": np_array_int([1131, 1132, 1133, 1134, 2131, 2132, 2133, 2134]),
            },
        },
        "bar": {
            "indptr": np.array([0, 4, 4, 8]),
            "data": {
                "x": np_array_int([1211, 1212, 1213, 1214, 2211, 2212, 2213, 2214]),
                "y": np_array_int([1221, 1222, 1223, 1224, 2221, 2222, 2223, 2224]),
                "z": np_array_int([1231, 1232, 1233, 1234, 2231, 2232, 2233, 2234]),
            },
        },
    }
    expected: BatchList = [
        {
            "foo": {
                "a": np_array_int([1111, 1112, 1113, 1114]),
                "b": np_array_int([1121, 1122, 1123, 1124]),
                "c": np_array_int([1131, 1132, 1133, 1134]),
            },
            "bar": {
                "x": np_array_int([1211, 1212, 1213, 1214]),
                "y": np_array_int([1221, 1222, 1223, 1224]),
                "z": np_array_int([1231, 1232, 1233, 1234]),
            },
        },
        {
            "foo": {
                "a": np_array_int([2111, 2112, 2113, 2114]),
                "b": np_array_int([2121, 2122, 2123, 2124]),
                "c": np_array_int([2131, 2132, 2133, 2134]),
            },
            "bar": {"x": np_array_int([]), "y": np_array_int([]), "z": np_array_int([])},
        },
        {
            "foo": {"a": np_array_int([]), "b": np_array_int([]), "c": np_array_int([])},
            "bar": {
                "x": np_array_int([2211, 2212, 2213, 2214]),
                "y": np_array_int([2221, 2222, 2223, 2224]),
                "z": np_array_int([2231, 2232, 2233, 2234]),
            },
        },
    ]
    actual = convert_batch_dataset_to_batch_list(update_data)
    assert_list_of_dicts_of_numpy_arrays_equal(expected, actual)


def test_get_and_verify_batch_sizes_inconsistent_batch_sizes_two_components():
    update_data: BatchDataset = {"foo": np.empty(shape=(3, 3)), "bar": np.empty(shape=(2, 3))}
    with pytest.raises(
        ValueError,
        match="Inconsistent number of batches in batch data. "
        "Component 'bar' contains 2 batches, while 'foo' contained 3 batches.",
    ):
        get_and_verify_batch_sizes(update_data)


def test_convert_get_and_verify_batch_sizes_inconsistent_batch_sizes_more_than_two_components():
    update_data: BatchDataset = {
        "foo": np.empty(shape=(3, 3)),
        "bar": np.empty(shape=(3, 3)),
        "baz": np.empty(shape=(2, 3)),
    }
    with pytest.raises(
        ValueError,
        match="Inconsistent number of batches in batch data. "
        "Component 'baz' contains 2 batches, while bar/foo contained 3 batches.",
    ):
        get_and_verify_batch_sizes(update_data)


@patch("power_grid_model._utils.get_and_verify_batch_sizes")
def test_convert_batch_dataset_to_batch_list_missing_key_sparse(_mock: MagicMock):
    update_data: BatchDataset = {"foo": {"a": np.empty(3), "data": np.empty(3)}}  # type: ignore
    with pytest.raises(KeyError, match="Invalid data for 'foo' component. Missing 'indptr' in sparse batch data. "):
        convert_batch_dataset_to_batch_list(update_data)


@patch("power_grid_model._utils.get_and_verify_batch_sizes")
def test_convert_batch_dataset_to_batch_list_invalid_type_sparse(_mock: MagicMock):
    update_data: BatchDataset = {"foo": "wrong type"}  # type: ignore
    with pytest.raises(
        TypeError,
        match="Invalid data for 'foo' component. " "Expecting a 1D/2D Numpy structured array or a dictionary of such.",
    ):
        convert_batch_dataset_to_batch_list(update_data)


DATA_FILTER_EVERYTHING = ComponentAttributeFilterOptions.everything
DATA_FILTER_RELEVANT = ComponentAttributeFilterOptions.relevant


@pytest.mark.parametrize(
    ("data_filter", "expected"),
    [
        (None, {CT.node: None, CT.sym_load: None, CT.source: None}),
        (
            DATA_FILTER_EVERYTHING,
            {CT.node: DATA_FILTER_EVERYTHING, CT.sym_load: DATA_FILTER_EVERYTHING, CT.source: DATA_FILTER_EVERYTHING},
        ),
        (
            DATA_FILTER_RELEVANT,
            {CT.node: DATA_FILTER_RELEVANT, CT.sym_load: DATA_FILTER_RELEVANT, CT.source: DATA_FILTER_RELEVANT},
        ),
        ([CT.node, CT.sym_load], {CT.node: None, CT.sym_load: None}),
        ({CT.node, CT.sym_load}, {CT.node: None, CT.sym_load: None}),
        ({CT.node: [], CT.sym_load: []}, {CT.node: [], CT.sym_load: []}),
        ({CT.node: [], CT.sym_load: ["p"]}, {CT.node: [], CT.sym_load: ["p"]}),
        ({CT.node: None, CT.sym_load: ["p"]}, {CT.node: None, CT.sym_load: ["p"]}),
        ({CT.node: DATA_FILTER_EVERYTHING, CT.sym_load: ["p"]}, {CT.node: DATA_FILTER_EVERYTHING, CT.sym_load: ["p"]}),
        ({CT.node: DATA_FILTER_RELEVANT, CT.sym_load: ["p"]}, {CT.node: DATA_FILTER_RELEVANT, CT.sym_load: ["p"]}),
        (
            {CT.node: DATA_FILTER_EVERYTHING, CT.sym_load: DATA_FILTER_EVERYTHING},
            {CT.node: DATA_FILTER_EVERYTHING, CT.sym_load: DATA_FILTER_EVERYTHING},
        ),
        (
            {CT.node: DATA_FILTER_RELEVANT, CT.sym_load: DATA_FILTER_RELEVANT},
            {CT.node: DATA_FILTER_RELEVANT, CT.sym_load: DATA_FILTER_RELEVANT},
        ),
        ({CT.node: ["u"], CT.sym_load: ["p"]}, {CT.node: ["u"], CT.sym_load: ["p"]}),
    ],
)
def test_process_data_filter(data_filter, expected):
    actual = process_data_filter(
        dataset_type=DT.sym_output,
        data_filter=data_filter,
        available_components=[CT.node, CT.sym_load, CT.source],
    )
    assert actual == expected


@pytest.mark.parametrize(
    ("data_filter", "available_components", "error", "match"),
    [
        ({"abc": 3, "def": None}, None, ValueError, "Invalid filter provided"),
        ({"abc": None, "def": None}, None, KeyError, "component types are unknown"),
        ({"abc": None, CT.sym_load: None}, None, KeyError, "component types are unknown"),
        ({"abc": ["xyz"], CT.sym_load: None}, None, KeyError, "component types are unknown"),
        ({CT.node: ["xyz"], CT.sym_load: None}, None, KeyError, "attributes are unknown"),
        ({CT.node: ["xyz1"], CT.sym_load: ["xyz2"]}, None, KeyError, "attributes are unknown"),
        ({CT.node: None, CT.sym_load: None}, [CT.node, "ghi"], KeyError, "component types are unknown"),
    ],
)
def test_process_data_filter__errors(data_filter, available_components, error, match):
    if available_components is None:
        available_components = [CT.node, CT.sym_load, CT.source]
    with pytest.raises(error, match=match):
        process_data_filter(
            dataset_type=DT.sym_output,
            data_filter=data_filter,
            available_components=available_components,
        )


def sample_output_data():
    output_data = {
        CT.node: initialize_array(DT.sym_output, CT.node, 4),
        CT.sym_load: initialize_array(DT.sym_output, CT.sym_load, 3),
        CT.source: initialize_array(DT.sym_output, CT.source, 1),
    }
    for comp in output_data:
        for attr in output_data[comp].dtype.names:
            output_data[comp][attr] = 0
    return output_data


@pytest.mark.parametrize(
    ("output_component_types", "expected"),
    [
        pytest.param(
            None,
            sample_output_data(),
            id="All row default",
        ),
        pytest.param(
            [CT.node, CT.sym_load],
            {k: v for k, v in sample_output_data().items() if k in [CT.node, CT.sym_load]},
            id="All row list filter",
        ),
        pytest.param(
            {CT.node, CT.sym_load},
            {k: v for k, v in sample_output_data().items() if k in [CT.node, CT.sym_load]},
            id="All row set filter",
        ),
        pytest.param([CT.shunt], {}, id="list/set filter-No component in filter present in data"),
        pytest.param(
            [CT.node, CT.shunt],
            {k: v for k, v in sample_output_data().items() if k in [CT.node]},
            id="list/set filter-Component in filter not present in data",
        ),
        pytest.param(
            {CT.node: [], CT.shunt: ["p"]}, {CT.node: dict()}, id="dict filter-Component in filter not present in data"
        ),
        pytest.param(
            {CT.node: [], CT.shunt: []},
            {CT.node: dict()},
            id="dict filter-Component in filter not present in data without attributes",
        ),
    ],
)
def test_copy_output_to_columnar_dataset(output_component_types, expected):
    actual = compatibility_convert_row_columnar_dataset(
        data=sample_output_data(),
        data_filter=output_component_types,
        dataset_type=DT.sym_output,
        available_components=list(sample_output_data().keys()),
    )
    assert actual.keys() == expected.keys()
    for comp_name in expected:
        np.testing.assert_array_equal(actual[comp_name], expected[comp_name])


@pytest.mark.parametrize(
    ("data", "dataset_type", "expected_size"),
    [
        pytest.param(np.empty(shape=(1,)), None, 1, id="row based single"),
        pytest.param(np.empty(shape=(3, 2)), None, 3, id="row based batch"),
        pytest.param({"u": np.empty(shape=(3,))}, DT.sym_output, 1, id="columnar single"),
        pytest.param({"u": np.empty(shape=(3, 2))}, DT.sym_output, 3, id="columnar batch"),
        pytest.param({"u": np.empty(shape=(4, 2, 3))}, DT.asym_output, 4, id="columnar asym batch"),
        pytest.param({"indptr": np.array([0, 1, 4]), "data": np.array([])}, None, 2, id="sparse data batch"),
        pytest.param({"indptr": np.array([0, 1]), "data": np.array([])}, None, 1, id="sparse data single"),
    ],
)
def test_get_batch_size(data, dataset_type, expected_size):
    assert get_batch_size(data, dataset_type, CT.node) == expected_size


@pytest.mark.parametrize(
    ("data", "dataset_type", "component", "error"),
    [
        pytest.param({"u": np.empty(shape=3)}, None, None, ValueError, id="No dataset type"),
        pytest.param({"u": np.empty(shape=3)}, DT.asym_output, None, ValueError, id="No component"),
        pytest.param(np.empty(shape=(1, 2, 3)), None, None, TypeError, id="Incorrect dimension for row based sym"),
        pytest.param(
            {"u": np.empty(shape=3)}, DT.asym_output, CT.node, TypeError, id="Incorrect dimension for columnar asym"
        ),
        pytest.param(
            {"u": np.empty(shape=(1, 2, 3))},
            DT.sym_output,
            CT.node,
            TypeError,
            id="Incorrect dimension for columnar sym",
        ),
    ],
)
def test_get_batch_size__errors(data, dataset_type, component, error):
    with pytest.raises(error):
        get_batch_size(data, dataset_type, component)


@pytest.mark.parametrize("dataset_type", [DT.input, DT.update, DT.sym_output, DT.asym_output, DT.sc_output])
def test_get_dataset_type(dataset_type):
    assert (
        get_dataset_type(
            data={
                CT.node: np.zeros(1, dtype=power_grid_meta_data[dataset_type]["node"]),
                CT.sym_load: np.zeros(1, dtype=power_grid_meta_data[dataset_type]["sym_load"]),
            }
        )
        == dataset_type
    )


def test_get_dataset_type__empty_data():
    with pytest.raises(ValueError):
        get_dataset_type(data={})


def test_get_dataset_type__conflicting_data():
    for first, second in itertools.product(
        [DT.input, DT.update, DT.sym_output, DT.asym_output, DT.sc_output],
        [DT.input, DT.update, DT.sym_output, DT.asym_output, DT.sc_output],
    ):
        data = {
            "node": np.zeros(1, dtype=power_grid_meta_data[first]["node"]),
            "sym_load": np.zeros(1, dtype=power_grid_meta_data[second]["sym_load"]),
        }
        if first == second:
            assert get_dataset_type(data=data) == first
        else:
            with pytest.raises(PowerGridError):
                get_dataset_type(data=data)


@pytest.fixture
def row_dense_data():
    source = initialize_array(DatasetType.update, ComponentType.source, (2, 3))
    source["id"] = [[0, 2, 3], [0, 2, 3]]
    source["u_ref"] = [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]]

    sym_load = initialize_array(DatasetType.update, ComponentType.sym_load, (2, 1))
    sym_load["id"] = [[1], [1]]
    sym_load["p_specified"] = [[100.0], [200.0]]

    return {
        ComponentType.source: source,
        ComponentType.sym_load: sym_load,
    }


@pytest.fixture
def row_sparse_data():
    transformer = initialize_array(DatasetType.update, ComponentType.transformer, 1)
    transformer["id"] = 1
    transformer["tap_pos"] = 3

    sym_gen = initialize_array(DatasetType.update, ComponentType.sym_gen, 8)
    sym_gen["id"] = [5, 6, 7, 8, 5, 6, 7, 8]
    sym_gen["q_specified"] = [1.1, 2.2, 3.3, 4.4, 4.4, 3.3, 2.2, 1.1]

    return {
        ComponentType.transformer: {
            "data": transformer,
            "indptr": np.array([0, 1, 1]),
        },
        ComponentType.sym_gen: {
            "data": sym_gen,
            "indptr": np.array([0, 4, 8]),
        },
    }


@pytest.fixture(params=["row_dense_data", "row_sparse_data"])
def row_data(request):
    return request.getfixturevalue(request.param)


def compare_row_data(actual_row_data, desired_row_data):
    assert actual_row_data.keys() == desired_row_data.keys()

    for comp_name in actual_row_data.keys():
        actual_component = actual_row_data[comp_name]
        desired_component = desired_row_data[comp_name]
        if is_sparse(actual_component):
            assert actual_component.keys() == desired_component.keys()
            assert np.array_equal(actual_component["indptr"], desired_component["indptr"])
            actual_component = actual_component["data"]
            desired_component = desired_component["data"]
        assert actual_component.dtype == desired_component.dtype
        assert actual_component.shape == desired_component.shape

        for field in actual_component.dtype.names:
            actual_attr = actual_component[field]
            desired_attr = desired_component[field]
            nan_a = np.isnan(actual_attr)
            nan_b = np.isnan(desired_attr)
            np.testing.assert_array_equal(nan_a, nan_b)
            np.testing.assert_allclose(actual_attr[~nan_a], desired_attr[~nan_b], rtol=1e-15)


def test_dense_row_data_to_from_col_data(row_data):
    # row data to columnar data and back
    col_data = compatibility_convert_row_columnar_dataset(
        row_data, ComponentAttributeFilterOptions.everything, DatasetType.update
    )
    new_row_data = compatibility_convert_row_columnar_dataset(col_data, None, DatasetType.update)
    compare_row_data(row_data, new_row_data)
